import inspect
import copy
from protorpc import messages
from .converters import converters as default_converters
from google.appengine.ext import ndb


class holder(object):
    pass


def _common_fields(entity, message, only=None, exclude=None):
    message_fields = [x.name for x in message.all_fields()]
    entity_properties = [k for k, v in entity._properties.iteritems()]

    if (inspect.isclass(entity) and not issubclass(entity, ndb.Expando)) and not isinstance(entity, ndb.Expando):
        fields = set(message_fields) & set(entity_properties)
    else:
        fields = set(message_fields) - set(['key'])

    if only:
        fields = set(only) & set(fields)

    if exclude:
        fields = [x for x in fields if x not in exclude]

    return message_fields, entity_properties, fields


def to_message(entity, message, converters=None, only=None, exclude=None, include_key=True):
    message_fields, entity_properties, fields = _common_fields(entity, message, only, exclude)

    converters = dict(default_converters.items() + converters.items()) if converters else default_converters

    # Key first
    values = {}

    # Key first
    if include_key:
        values['datastore_key'] = converters['Key'].to_message(entity, 'key', 'datastore_key', entity.key) if entity.key else None

    # Other fields
    for field in fields:
        if field not in entity._properties:
            continue

        property = entity._properties[field]
        message_field = message.field_by_name(field)
        value = getattr(entity, field)

        converter = converters[property.__class__.__name__]

        if converter:
            if value is not None:  # only try to convert if the value is meaningful, otherwise leave it as Falsy.
                if property._repeated:
                    value = [converter.to_message(entity, property, message_field, x) if x else x for x in value]
                else:
                    value = converter.to_message(entity, property, message_field, value)
            values[field] = value

    if inspect.isclass(message):
        return message(**values)
    else:
        for name, value in values.iteritems():
            setattr(message, name, value)
        return message


def to_entity(message, model, converters=None, only=None, exclude=None, include_key=True):
    message_fields, entity_properties, fields = _common_fields(model, message, only, exclude)

    converters = dict(default_converters.items() + converters.items()) if converters else default_converters

    values = {}

    # Key first, if it's there
    if include_key and hasattr(message, 'datastore_key') and message.datastore_key:
        values['key'] = converters['Key'].to_model(messages, 'key', 'datastore_key', message.datastore_key)

    # Other fields
    for field in fields:
        if field in model._properties:
            property = model._properties[field]
        elif (inspect.isclass(model) and issubclass(model, ndb.Expando)) or isinstance(model, ndb.Expando):
            property = ndb.GenericProperty(field)
        else:
            continue

        converter = converters[property.__class__.__name__]
        message_field = message.field_by_name(field)
        value = getattr(message, field)

        if converter:
            if value is not None:
                if property._repeated:
                    value = [converter.to_model(message, property, message_field, x) if x else x for x in value]
                else:
                    value = converter.to_model(message, property, message_field, value)
            elif property._repeated:
                value = []

            values[field] = value

    if inspect.isclass(model):
        return model(**values)
    else:
        model.populate(**values)
        return model


def model_message(Model, only=None, exclude=None, converters=None, include_key=True):
    class_name = Model.__name__ + 'Message'

    props = Model._properties
    sorted_props = sorted(props.iteritems(), key=lambda prop: prop[1]._creation_counter)
    field_names = [x[0] for x in sorted_props if x[0]]

    if exclude:
        field_names = [x for x in field_names if x not in exclude]

    if only:
        field_names = [x for x in field_names if x in only]

    converters = dict(default_converters.items() + converters.items()) if converters else default_converters

    # Add in the key field.
    key_holder = holder()
    key_holder.name = 'key',
    key_holder._repeated = False
    field_dict = {
        'datastore_key': converters['Key'].to_field(Model, key_holder, 1)
    }

    # Add all other fields.
    for count, name in enumerate(field_names, start=2):
        prop = props[name]
        converter = converters.get(prop.__class__.__name__, None)

        if converter:
            field_dict[name] = converter.to_field(Model, prop, count)

    return type(class_name, (messages.Message,), field_dict)


def list_message(message_type):
    name = message_type.__name__ + 'List'
    fields = {
        'items': messages.MessageField(message_type, 1, repeated=True),
        'next_page_token': messages.StringField(2)
    }
    return type(name, (messages.Message,), fields)


def compose(*args):
    fields = {}
    name = 'Composed'

    for message_cls in args:
        name += message_cls.__name__
        for field in message_cls.all_fields():
            fields[field.name] = field

    for n, orig_field in enumerate(fields.values(), 1):
        field = copy.copy(orig_field)
        # This is so ridiculously hacky. I'm not proud of it, but the alternative to doing this is trying to reconstruct each
        # field by figuring out the arguments originally passed into __init__. I think this is honestly a little cleaner.
        object.__setattr__(field, 'number', n)
        fields[field.name] = field

    return type(name, (messages.Message,), fields)



import datetime
from protorpc import messages, message_types, util
from google.appengine.ext import ndb
from google.appengine.api import users
from .types import DateMessage, TimeMessage, UserMessage, KeyMessage, GeoPtMessage


class Converter(object):
    @staticmethod
    def to_message(Mode, property, field, value):
        return value

    @staticmethod
    def to_model(Message, property, field, value):
        return value

    @staticmethod
    def to_field(Model, property, count):
        return None


class StringConverter(Converter):
    @staticmethod
    def to_field(Model, property, count):
        return messages.StringField(count, repeated=property._repeated, required=property._required)


class BytesConverter(Converter):
    @staticmethod
    def to_field(Model, property, count):
        return messages.BytesField(count, repeated=property._repeated)


class BooleanConverter(Converter):
    @staticmethod
    def to_field(Model, property, count):
        return messages.BooleanField(count, repeated=property._repeated)


class IntegerConverter(Converter):
    @staticmethod
    def to_field(Model, property, count):
        return messages.IntegerField(count, repeated=property._repeated)


class FloatConverter(Converter):
    @staticmethod
    def to_field(Model, property, count):
        return messages.FloatField(count, repeated=property._repeated)


class DateTimeConverter(Converter):
    @staticmethod
    def to_field(Model, property, count):
        return message_types.DateTimeField(count, repeated=property._repeated)


class DateConverter(Converter):
    @staticmethod
    def to_message(Mode, property, field, value):
        return DateMessage(
            year=value.year,
            month=value.month,
            day=value.day)

    @staticmethod
    def to_model(Message, property, field, value):
        return datetime.date(value.year, value.month, value.day)

    @staticmethod
    def to_field(Model, property, count):
        return messages.MessageField(DateMessage, count, repeated=property._repeated)


class TimeConverter(Converter):
    @staticmethod
    def to_message(Mode, property, field, value):
        time_zone_offset = 0
        if value.tzinfo is not None:
            utc_offset = value.tzinfo.utcoffset(value)
            if utc_offset is not None:
                time_zone_offset = int(utc_offset.total_seconds() / 60)

        return TimeMessage(
            hour=value.hour,
            minute=value.minute,
            second=value.second,
            microsecond=value.microsecond,
            time_zone_offset=time_zone_offset)

    @staticmethod
    def to_model(Message, property, field, value):
        timezone = None
        if value.time_zone_offset:
            timezone = util.TimeZoneOffset(value.time_zone_offset)
        return datetime.time(value.hour, value.minute, value.second, value.microsecond, timezone)

    @staticmethod
    def to_field(Model, property, count):
        return messages.MessageField(TimeMessage, count, repeated=property._repeated)


class UserConverter(Converter):
    @staticmethod
    def to_message(Mode, property, field, value):
        return UserMessage(
            email=value.email(),
            user_id=value.user_id(),
            nickname=value.nickname())

    @staticmethod
    def to_model(Message, property, field, value):
        if isinstance(value, basestring):
            return users.User(email=value)
        elif isinstance(value, UserMessage) and value.email:
            return users.User(email=value.email)

    @staticmethod
    def to_field(Model, property, count):
        return messages.MessageField(UserMessage, count, repeated=property._repeated)


class BlobKeyConverter(Converter):
    @staticmethod
    def to_message(Mode, property, field, value):
        return str(value)

    @staticmethod
    def to_model(Message, property, field, value):
        return ndb.BlobKey(value)

    @staticmethod
    def to_field(Model, property, count):
        return messages.StringField(count, repeated=property._repeated)


class KeyConverter(Converter):
    @staticmethod
    def to_message(Mode, property, field, value):
        return KeyMessage(
            urlsafe=value.urlsafe(),
            id=str(value.integer_id()) if value.integer_id() else value.string_id().decode('utf8'),
            kind=value.kind())

    @staticmethod
    def to_model(Message, property, field, value):
        if isinstance(value, basestring):
            return ndb.Key(urlsafe=value)
        elif isinstance(value, KeyMessage):
            return ndb.Key(urlsafe=value.urlsafe)

    @staticmethod
    def to_field(Model, property, count):
        return messages.MessageField(KeyMessage, count, repeated=property._repeated)


class GeoPtConverter(Converter):
    @staticmethod
    def to_message(Mode, property, field, value):
        return GeoPtMessage(lat=value.lat, lon=value.lon)

    @staticmethod
    def to_model(Message, property, field, value):
        if value:
            return ndb.GeoPt(value.lat, value.lon)

    @staticmethod
    def to_field(Model, property, count):
        return messages.MessageField(GeoPtMessage, count, repeated=property._repeated)


class StructuredConverter(Converter):
    @staticmethod
    def to_message(Model, property, field, value):
        from .translators import to_message
        return to_message(value, field.type)

    @staticmethod
    def to_model(Message, property, field, value):
        from .translators import to_entity
        if value:
            return to_entity(value, property._modelclass)

    @staticmethod
    def to_field(Model, property, count):
        from .translators import model_message

        message_class = model_message(property._modelclass)
        return messages.MessageField(message_class, count, repeated=property._repeated)

converters = {
    'Key': KeyConverter,
    'BooleanProperty': BooleanConverter,
    'IntegerProperty': IntegerConverter,
    'FloatProperty': FloatConverter,
    'BlobProperty': BytesConverter,
    'StringProperty': StringConverter,
    'TextProperty': StringConverter,
    'DateTimeProperty': DateTimeConverter,
    'TimeProperty': TimeConverter,
    'DateProperty': DateConverter,
    'UserProperty': UserConverter,
    'GeoPtProperty': GeoPtConverter,
    'KeyProperty': KeyConverter,
    'BlobKeyProperty': BlobKeyConverter,
    'StructuredProperty': StructuredConverter,
    'LocalStructuredProperty': StructuredConverter
}



from setuptools import setup

VERSION = '1.1.1'

setup(
    name="protopigeon",
    version=VERSION,
    author='Jon Parrott',
    author_email='jjramone13@gmail.com',
    maintainer='Jon Parrott / Cloud Sherpas',
    maintainer_email='jon.wayne.parrott@gmail.com',
    description="A helper library for working with Google's protorpc and App Engine's datastore",
    url='https://github.com/jonparrott/Protopigeon',
    license='Apache License 2.0',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
    ],
    packages=['protopigeon'],
    install_requires=[
    ],
)



from protorpc import messages


class TimeMessage(messages.Message):
    hour = messages.IntegerField(1)
    minute = messages.IntegerField(2)
    second = messages.IntegerField(3)
    microsecond = messages.IntegerField(4)
    time_zone_offset = messages.IntegerField(5)


class DateMessage(messages.Message):
    year = messages.IntegerField(1)
    month = messages.IntegerField(2)
    day = messages.IntegerField(3)


class UserMessage(messages.Message):
    email = messages.StringField(1)
    user_id = messages.StringField(2)
    nickname = messages.StringField(3)


class KeyMessage(messages.Message):
    urlsafe = messages.StringField(1)
    id = messages.StringField(2)
    kind = messages.StringField(4)


class GeoPtMessage(messages.Message):
    lat = messages.FloatField(1)
    lon = messages.FloatField(2)
